/*
 * PolicyFactory.cc
 *
 * Copyright (c) 2003, 2004 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file PolicyFactory.cc
 *
 * The shared library interface for libpolicies.
 */

#include "config.h"

#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#include <iostream>

#include <assert_pp.h>
#include <HeyParser.hh>
#include <factory_library.h>

#include "ManagerImpl.hh"
#include "PartitionPolicyImpl.hh"
#include "StrictPolicyImpl.hh"
#include "WeightedPolicyImpl.hh"

using namespace std;

/**
 * Handle 'hey's directed at the library.
 *
 * @param orb Pointer to the ORB.
 * @param poa The POA to use when activating objects.
 * @param hp The HeyParser that describes the action to take.
 * @param out Standard out.
 * @param err Standard error.
 * @return Zero on success, an errno value otherwise.
 */
static int pfHey(CORBA::ORB_ptr orb,
		 PortableServer::POA_ptr poa,
		 HeyParser &hp,
		 ostream &out,
		 ostream &err)
{
    static HeyPropertyInfo suites[] = {
	HeyPropertyInfo("strict-policy",
			(1L << HeyParser::CREATE_PROPERTY),
			"",
			"name:string - The name of the StrictPolicy.\n"
			"\n"
			"Create a StrictPolicy object.  This policy resolves "
			"contention by completely\n"
			"satisfying task requests in order of priority until "
			"there is no more CPU\n"
			"available.\n"),
	HeyPropertyInfo("weighted-policy",
			(1L << HeyParser::CREATE_PROPERTY),
			"",
			"name:string - The name of the WeightedPolicy.\n"
			"\n"
			"Create a WeightedPolicy object.  This policy "
			"resolves contention using a\n"
			"proportional share scheme.  For example, a task with "
			"a weight of 20 will get\n"
			"twice as much CPU as one with 10.\n"),
	HeyPropertyInfo("partition-policy",
			(1L << HeyParser::CREATE_PROPERTY),
			"",
			"name:string - The name of the PartitionPolicy.\n"
			"\n"
			"Create a PartitionPolicy object.  This policy acts "
			"as a meta-policy that\n"
			"partitions tasks into separate sub-policies, which "
			"can of any type.\n"),
	HeyPropertyInfo::HPI_NULL
    };

    const char *prop_name, *prop_value;
    int retval = EXIT_FAILURE;

    require(!CORBA::is_nil(orb));
    require(!CORBA::is_nil(poa));

    switch( hp.what() )
    {
    case HeyParser::LIST_PROPERTIES:
	try
	{
	    hp.popProperty(prop_name, prop_value);
	}
	catch(const HeyParserException &e)
	{
	    out << "strict-policy" << endl
		<< "weighted-policy" << endl
		<< "partition-policy" << endl;
	}
	break;
    case HeyParser::CREATE_PROPERTY:
	hp.popProperty(prop_name, prop_value);
	if( strcasecmp(prop_name, "strict-policy") == 0 )
	{
	    StrictPolicyImpl *spi = new StrictPolicyImpl(hp.withValue("name"));
	    BrokerPolicies::StrictPolicy_var sp;
	    PortableServer::ObjectId_var oid;
	    
	    oid = poa->activate_object(spi);
	    sp = BrokerPolicies::StrictPolicy::
		_narrow(poa->id_to_reference(oid.in()));
	    out << orb->object_to_string(sp._retn()) << endl;
	    retval = EXIT_SUCCESS;
	}
	else if( strcasecmp(prop_name, "weighted-policy") == 0 )
	{
	    BrokerPolicies::WeightedPolicy_var wp;
	    PortableServer::ObjectId_var oid;
	    WeightedPolicyImpl *wpi;

	    wpi = new WeightedPolicyImpl(hp.withValue("name"));
	    oid = poa->activate_object(wpi);
	    wp = BrokerPolicies::WeightedPolicy::
		_narrow(poa->id_to_reference(oid.in()));
	    out << orb->object_to_string(wp._retn()) << endl;
	    retval = EXIT_SUCCESS;
	}
	else if( strcasecmp(prop_name, "partition-policy") == 0 )
	{
	    BrokerPolicies::PartitionPolicy_var pp;
	    PortableServer::ObjectId_var oid;
	    PartitionPolicyImpl *ppi;

	    ppi = new PartitionPolicyImpl(hp.withValue("name"));
	    oid = poa->activate_object(ppi);
	    pp = BrokerPolicies::PartitionPolicy::
		_narrow(poa->id_to_reference(oid.in()));
	    out << orb->object_to_string(pp._retn()) << endl;
	    retval = EXIT_SUCCESS;
	}
	else
	{
	    err << "Unknown property: " << prop_name << endl;
	}
	break;
    case HeyParser::GET_SUITES:
	try
	{
	    hp.popProperty(prop_name, prop_value);
	    err << "Unknown property: " << prop_name << endl;
	}
	catch(const HeyParserException &e)
	{
	    out << suites;
	    retval = EXIT_SUCCESS;
	}
	break;
    default:
	err << "Unknown verb" << endl;
	break;
    }
    return( retval );
}

/**
 * Handle 'hey's directed at a StrictPolicy object.
 *
 * @param orb Pointer to the ORB.
 * @param hp The HeyParser that describes the action to take.
 * @return Zero on success, an errno value otherwise.
 */
static int pfStrictPolicyHey(CORBA::ORB_ptr orb, HeyParser &hp)
{
    static HeyPropertyInfo suites[] = {
	HeyPropertyInfo("tasks",
			(1L << HeyParser::LIST_PROPERTIES),
			"",
			"List tasks currently being managed.\n"),
	HeyPropertyInfo("tasks",
			(1L << HeyParser::GET_SUITES),
			"",
			"Get a description of the sub-properties of task.\n"),
	HeyPropertyInfo("task",
			(1L << HeyParser::GET_PROPERTY),
			"name:string - The task name.",
			"Get the IOR for the task matching the given name.\n"),
	HeyPropertyInfo("max-cpu",
			(1L << HeyParser::SET_PROPERTY) |
			(1L << HeyParser::GET_PROPERTY),
			"",
			"Get or set the maximum CPU allocation.  The value is "
			"a floating point number\n"
			"between 0.0 and 1.0 that represents the percentage "
			"of CPU that the policy\n"
			"will allocate to the tasks it manages.\n"),
	HeyPropertyInfo::HPI_NULL
    };
    
    static HeyPropertyInfo task_suites[] = {
	HeyPropertyInfo("priority",
			(1L << HeyParser::SET_PROPERTY) |
			(1L << HeyParser::GET_PROPERTY),
			"",
			"Set or get the priority of the given task in this "
			"policy.\n"),
	HeyPropertyInfo::HPI_NULL
    };
    
    const char *prop_name, *prop_value, *to_value;
    BrokerPolicies::StrictPolicy_var sp;
    int retval = EXIT_FAILURE;
    Broker::TaskList_var tl;
    CORBA::Object_var obj;
    Broker::Task_var task;
    
    obj = orb->string_to_object(hp.who());
    sp = BrokerPolicies::StrictPolicy::_narrow(obj.in());
    if( CORBA::is_nil(sp.in()) )
    {
	throw CORBA::BAD_PARAM();
    }
    
    tl = sp->GetTaskList();
    switch( hp.what() )
    {
    case HeyParser::LIST_PROPERTIES:
	try
	{
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "tasks") == 0 )
	    {
		unsigned int lpc;
		
		for( lpc = 0; lpc < tl->length(); lpc++ )
		{
		    cout << tl[lpc]->Name() << endl;
		}
		retval = EXIT_SUCCESS;
	    }
	    else if( strcasecmp(prop_name, "task") == 0 )
	    {
		cout << "priority" << endl;
	    }
	    else
	    {
		cerr << "Unknown property: " << prop_name << endl;
	    }
	}
	catch(const HeyParserException &e)
	{
	    cout << "max-cpu" << endl;
	    cout << "tasks" << endl;
	    cout << "task" << endl;
	    retval = EXIT_SUCCESS;
	}
	break;
    case HeyParser::GET_PROPERTY:
	hp.popProperty(prop_name, prop_value);
	if( strcasecmp(prop_name, "task") == 0 )
	{
	    task = ManagerImpl::ResolveTask(orb, tl, prop_value);
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "priority") == 0 )
	    {
		cout << sp->GetTaskPriority(task.in()) << endl;
		retval = EXIT_SUCCESS;
	    }
	    else
	    {
		cout << orb->object_to_string(task.in()) << endl;
		retval = EXIT_SUCCESS;
	    }
	}
	else if( strcasecmp(prop_name, "max-cpu") == 0 )
	{
	    cout << sp->GetMaxCPUAllocation() << endl;
	    retval = EXIT_SUCCESS;
	}
	else
	{
	    cerr << "Unknown property: " << prop_name << endl;
	}
	break;
    case HeyParser::SET_PROPERTY:
	hp.popProperty(prop_name, prop_value);
	if( strcasecmp(prop_name, "task") == 0 )
	{
	    task = ManagerImpl::ResolveTask(orb, tl, prop_value);
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "priority") == 0 )
	    {
		CORBA::Short pri;
		
		to_value = hp.to();
		if( sscanf(to_value, "%hd", &pri) == 1 )
		{
		    sp->SetTaskPriority(task.in(), pri);
		    cout << "ok" << endl;
		    retval = EXIT_SUCCESS;
		}
		else
		{
		    cerr << "Priority not a number: "
			 << to_value
			 << endl;
		}
	    }
	    else
	    {
		cerr << "Unknown property: " << prop_name << endl;
	    }
	}
	else if( strcasecmp(prop_name, "max-cpu") == 0 )
	{
	    CORBA::Float pct;
	    
	    to_value = hp.to();
	    if( sscanf(to_value, "%f", &pct) == 1 )
	    {
		if( (pct < BrokerPolicies::StrictPolicy::CPU_ALLOCATION_MIN) ||
		    (pct > BrokerPolicies::StrictPolicy::CPU_ALLOCATION_MAX) )
		{
		    cerr << "Invalid percentage, expecting a "
			 << "value greater than "
			 << BrokerPolicies::StrictPolicy::CPU_ALLOCATION_MIN
			 << " and less than "
			 << BrokerPolicies::StrictPolicy::CPU_ALLOCATION_MAX
			 << endl;
		}
		else
		{
		    sp->SetMaxCPUAllocation(pct);
		    cout << "ok" << endl;
		    retval = EXIT_SUCCESS;
		}
	    }
	    else
	    {
		cerr << "max-cpu value not a percentage: "
		     << to_value
		     << endl;
	    }
	}
	else
	{
	    cerr << "Unknown property: " << prop_name << endl;
	}
	break;
    case HeyParser::GET_SUITES:
	try
	{
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "task") == 0 )
	    {
		cout << task_suites;
		retval = EXIT_SUCCESS;
	    }
	    else
	    {
		cerr << "Unknown property: " << prop_name << endl;
	    }
	}
	catch(const HeyParserException &e)
	{
	    cout << suites;
	    retval = EXIT_SUCCESS;
	}
	break;
    default:
	cerr << "Unknown verb..." << endl;
	break;
    }
    return( retval );
}

/**
 * Handle 'hey's directed at a WeightedPolicy object.
 *
 * @param orb Pointer to the ORB.
 * @param hp The HeyParser that describes the action to take.
 * @return Zero on success, an errno value otherwise.
 */
static int pfWeightedPolicyHey(CORBA::ORB_ptr orb, HeyParser &hp)
{
    static HeyPropertyInfo suites[] = {
	HeyPropertyInfo("tasks",
			(1L << HeyParser::LIST_PROPERTIES),
			"",
			"List tasks currently being managed.\n"),
	HeyPropertyInfo("tasks",
			(1L << HeyParser::GET_SUITES),
			"",
			"Get a description of the sub-properties of task.\n"),
	HeyPropertyInfo("task",
			(1L << HeyParser::GET_PROPERTY),
			"name:string - The task name.",
			"Get the IOR for the task matching the given name.\n"),
	HeyPropertyInfo("max-cpu",
			(1L << HeyParser::SET_PROPERTY) |
			(1L << HeyParser::GET_PROPERTY),
			"",
			"Get or set the maximum CPU allocation.  The value is "
			"a floating point number\n"
			"between 0.0 and 1.0 that represents the percentage "
			"of CPU that the policy\n"
			"will allocate to the tasks it manages.\n"),
	HeyPropertyInfo::HPI_NULL
    };
    
    static HeyPropertyInfo task_suites[] = {
	HeyPropertyInfo("weight",
			(1L << HeyParser::SET_PROPERTY) |
			(1L << HeyParser::GET_PROPERTY),
			"",
			"Set or get the weight of the given task in this "
			"policy.\n"),
	HeyPropertyInfo::HPI_NULL
    };
    
    const char *prop_name, *prop_value, *to_value;
    BrokerPolicies::WeightedPolicy_var sp;
    int retval = EXIT_FAILURE;
    Broker::TaskList_var tl;
    CORBA::Object_var obj;
    Broker::Task_var task;
    
    obj = orb->string_to_object(hp.who());
    sp = BrokerPolicies::WeightedPolicy::_narrow(obj.in());
    if( CORBA::is_nil(sp.in()) )
    {
	throw CORBA::BAD_PARAM();
    }
    
    tl = sp->GetTaskList();
    switch( hp.what() )
    {
    case HeyParser::LIST_PROPERTIES:
	try
	{
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "tasks") == 0 )
	    {
		unsigned int lpc;
		
		for( lpc = 0; lpc < tl->length(); lpc++ )
		{
		    cout << tl[lpc]->Name() << endl;
		}
		retval = EXIT_SUCCESS;
	    }
	    else if( strcasecmp(prop_name, "task") == 0 )
	    {
		cout << "weight" << endl;
	    }
	    else
	    {
		cerr << "Unknown property: " << prop_name << endl;
	    }
	}
	catch(const HeyParserException &e)
	{
	    cout << "max-cpu" << endl;
	    cout << "tasks" << endl;
	    cout << "task" << endl;
	    retval = EXIT_SUCCESS;
	}
	break;
    case HeyParser::GET_PROPERTY:
	hp.popProperty(prop_name, prop_value);
	if( strcasecmp(prop_name, "task") == 0 )
	{
	    task = ManagerImpl::ResolveTask(orb, tl, prop_value);
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "weight") == 0 )
	    {
		cout << sp->GetTaskWeight(task.in()) << endl;
		retval = EXIT_SUCCESS;
	    }
	    else
	    {
		cout << orb->object_to_string(task.in()) << endl;
		retval = EXIT_SUCCESS;
	    }
	}
	else if( strcasecmp(prop_name, "max-cpu") == 0 )
	{
	    cout << sp->GetMaxCPUAllocation() << endl;
	    retval = EXIT_SUCCESS;
	}
	else
	{
	    cerr << "Unknown property: " << prop_name << endl;
	}
	break;
    case HeyParser::SET_PROPERTY:
	hp.popProperty(prop_name, prop_value);
	if( strcasecmp(prop_name, "task") == 0 )
	{
	    task = ManagerImpl::ResolveTask(orb, tl, prop_value);
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "weight") == 0 )
	    {
		CORBA::Short pri;
		
		to_value = hp.to();
		if( sscanf(to_value, "%hd", &pri) == 1 )
		{
		    sp->SetTaskWeight(task.in(), pri);
		    cout << "ok" << endl;
		    retval = EXIT_SUCCESS;
		}
		else
		{
		    cerr << "Weight not a number: "
			 << to_value
			 << endl;
		}
	    }
	    else
	    {
		cerr << "Unknown property: " << prop_name << endl;
	    }
	}
	else if( strcasecmp(prop_name, "max-cpu") == 0 )
	{
	    CORBA::Float pct;
	    
	    to_value = hp.to();
	    if( sscanf(to_value, "%f", &pct) == 1 )
	    {
		if( (pct < BrokerPolicies::WeightedPolicy::CPU_ALLOCATION_MIN) ||
		    (pct > BrokerPolicies::WeightedPolicy::CPU_ALLOCATION_MAX) )
		{
		    cerr << "Invalid percentage, expecting a "
			 << "value greater than "
			 << BrokerPolicies::WeightedPolicy::CPU_ALLOCATION_MIN
			 << " and less than "
			 << BrokerPolicies::WeightedPolicy::CPU_ALLOCATION_MAX
			 << endl;
		}
		else
		{
		    sp->SetMaxCPUAllocation(pct);
		    cout << "ok" << endl;
		    retval = EXIT_SUCCESS;
		}
	    }
	    else
	    {
		cerr << "max-cpu value not a percentage: "
		     << to_value
		     << endl;
	    }
	}
	else
	{
	    cerr << "Unknown property: " << prop_name << endl;
	}
	break;
    case HeyParser::GET_SUITES:
	try
	{
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "task") == 0 )
	    {
		cout << task_suites;
		retval = EXIT_SUCCESS;
	    }
	    else
	    {
		cerr << "Unknown property: " << prop_name << endl;
	    }
	}
	catch(const HeyParserException &e)
	{
	    cout << suites;
	    retval = EXIT_SUCCESS;
	}
	break;
    default:
	cerr << "Unknown verb..." << endl;
	break;
    }
    return( retval );
}

/**
 * Handle 'hey's directed at a PartitionPolicy object.
 *
 * @param orb Pointer to the ORB.
 * @param hp The HeyParser that describes the action to take.
 * @return Zero on success, an errno value otherwise.
 */
static int pfPartitionPolicyHey(CORBA::ORB_ptr orb, HeyParser &hp)
{
    static HeyPropertyInfo suites[] = {
	HeyPropertyInfo("tasks",
			(1L << HeyParser::LIST_PROPERTIES),
			"",
			"List tasks currently being managed.\n"),
	HeyPropertyInfo("tasks",
			(1L << HeyParser::GET_SUITES),
			"",
			"Get a description of the sub-properties of task.\n"),
	HeyPropertyInfo("task",
			(1L << HeyParser::GET_PROPERTY),
			"name:string - The task name.",
			"Get the IOR for the task matching the given name.\n"),
	HeyPropertyInfo("max-cpu",
			(1L << HeyParser::SET_PROPERTY) |
			(1L << HeyParser::GET_PROPERTY),
			"",
			"Get or set the maximum CPU allocation.  The value is "
			"a floating point number\n"
			"between 0.0 and 1.0 that represents the percentage "
			"of CPU that the policy\n"
			"will allocate to the tasks it manages.\n"),
	HeyPropertyInfo::HPI_NULL
    };
    
    static HeyPropertyInfo task_suites[] = {
	HeyPropertyInfo("policy",
			(1L << HeyParser::SET_PROPERTY) |
			(1L << HeyParser::GET_PROPERTY),
			"",
			"Set or get the policy for the given task.\n"),
	HeyPropertyInfo::HPI_NULL
    };
    
    const char *prop_name, *prop_value, *to_value;
    BrokerPolicies::PartitionPolicy_var sp;
    int retval = EXIT_FAILURE;
    Broker::TaskList_var tl;
    CORBA::Object_var obj;
    Broker::Task_var task;
    
    obj = orb->string_to_object(hp.who());
    sp = BrokerPolicies::PartitionPolicy::_narrow(obj.in());
    if( CORBA::is_nil(sp.in()) )
    {
	throw CORBA::BAD_PARAM();
    }
    
    tl = sp->GetTaskList();
    switch( hp.what() )
    {
    case HeyParser::LIST_PROPERTIES:
	try
	{
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "tasks") == 0 )
	    {
		unsigned int lpc;
		
		for( lpc = 0; lpc < tl->length(); lpc++ )
		{
		    cout << tl[lpc]->Name() << endl;
		}
		retval = EXIT_SUCCESS;
	    }
	    else if( strcasecmp(prop_name, "task") == 0 )
	    {
		cout << "weight" << endl;
	    }
	    else
	    {
		cerr << "Unknown property: " << prop_name << endl;
	    }
	}
	catch(const HeyParserException &e)
	{
	    cout << "max-cpu" << endl;
	    cout << "tasks" << endl;
	    cout << "task" << endl;
	    retval = EXIT_SUCCESS;
	}
	break;
    case HeyParser::GET_PROPERTY:
	hp.popProperty(prop_name, prop_value);
	if( strcasecmp(prop_name, "task") == 0 )
	{
	    task = ManagerImpl::ResolveTask(orb, tl, prop_value);
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "policy") == 0 )
	    {
		Broker::Policy_var policy;

		policy = sp->GetTaskPolicy(task.in());
		cout << orb->object_to_string(policy.in()) << endl;
		retval = EXIT_SUCCESS;
	    }
	    else
	    {
		cout << orb->object_to_string(task.in()) << endl;
		retval = EXIT_SUCCESS;
	    }
	}
	else if( strcasecmp(prop_name, "default-policy") == 0 )
	{
	    Broker::Policy_var policy;

	    policy = sp->DefaultPolicy();
	    cout << orb->object_to_string(policy.in()) << endl;
	}
	else if( strcasecmp(prop_name, "max-cpu") == 0 )
	{
	    cout << sp->GetMaxCPUAllocation() << endl;
	    retval = EXIT_SUCCESS;
	}
	else
	{
	    cerr << "Unknown property: " << prop_name << endl;
	}
	break;
    case HeyParser::SET_PROPERTY:
	hp.popProperty(prop_name, prop_value);
	if( strcasecmp(prop_name, "task") == 0 )
	{
	    task = ManagerImpl::ResolveTask(orb, tl, prop_value);
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "policy") == 0 )
	    {
		Broker::Policy_var policy;
		CORBA::Object_var obj;

		obj = orb->string_to_object(hp.to());
		if( !CORBA::is_nil(obj.in()) )
		{
		    policy = Broker::Policy::_narrow(obj.in());
		    if( CORBA::is_nil(policy.in()) )
		    {
			cerr << "Error: Policy is not an object." << endl;
			throw CORBA::BAD_PARAM();
		    }
		}
		
		sp->SetTaskPolicy(task.in(), policy.in());
		cout << "ok" << endl;
	    }
	    else
	    {
		cerr << "Unknown property: " << prop_name << endl;
	    }
	}
	else if( strcasecmp(prop_name, "default-policy") == 0 )
	{
	    Broker::Policy_var policy;
	    CORBA::Object_var obj;
	    
	    obj = orb->string_to_object(hp.to());
	    if( !CORBA::is_nil(obj.in()) )
	    {
		policy = Broker::Policy::_narrow(obj.in());
		if( CORBA::is_nil(policy.in()) )
		{
		    cerr << "Error: Policy is not an object." << endl;
		    throw CORBA::BAD_PARAM();
		}
	    }
	    
	    sp->DefaultPolicy(policy.in());
	    cout << "ok" << endl;
	}
	else if( strcasecmp(prop_name, "max-cpu") == 0 )
	{
	    CORBA::Float pct;
	    
	    to_value = hp.to();
	    if( sscanf(to_value, "%f", &pct) == 1 )
	    {
		if( (pct < BrokerPolicies::PartitionPolicy::CPU_ALLOCATION_MIN) ||
		    (pct > BrokerPolicies::PartitionPolicy::CPU_ALLOCATION_MAX) )
		{
		    cerr << "Invalid percentage, expecting a "
			 << "value greater than "
			 << BrokerPolicies::PartitionPolicy::CPU_ALLOCATION_MIN
			 << " and less than "
			 << BrokerPolicies::PartitionPolicy::CPU_ALLOCATION_MAX
			 << endl;
		}
		else
		{
		    sp->SetMaxCPUAllocation(pct);
		    cout << "ok" << endl;
		    retval = EXIT_SUCCESS;
		}
	    }
	    else
	    {
		cerr << "max-cpu value not a percentage: "
		     << to_value
		     << endl;
	    }
	}
	else
	{
	    cerr << "Unknown property: " << prop_name << endl;
	}
	break;
    case HeyParser::GET_SUITES:
	try
	{
	    hp.popProperty(prop_name, prop_value);
	    if( strcasecmp(prop_name, "task") == 0 )
	    {
		cout << task_suites;
		retval = EXIT_SUCCESS;
	    }
	    else
	    {
		cerr << "Unknown property: " << prop_name << endl;
	    }
	}
	catch(const HeyParserException &e)
	{
	    cout << suites;
	    retval = EXIT_SUCCESS;
	}
	break;
    default:
	cerr << "Unknown verb..." << endl;
	break;
    }
    return( retval );
}

int FACTORY_METHOD_SYMBOL(factory_library_op_t op, int tag, va_list args)
{
    int retval = EXIT_FAILURE;

    struct {
	const char *hey_server_hint;
	HeyParser *hey_parser;
	CORBA::ORB_ptr orb;
	PortableServer::POA_ptr poa;
	ostream *out;
	ostream *err;
    } attr = {
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    };
    
    while( tag != FMA_TAG_DONE )
    {
	switch( tag )
	{
	case FMA_ORB:
	    attr.orb = va_arg(args, CORBA::ORB_ptr);
	    break;
	case FMA_POA:
	    attr.poa = va_arg(args, PortableServer::POA_ptr);
	    break;
	case FMA_HeyParser:
	    attr.hey_parser = va_arg(args, HeyParser *);
	    break;
	case FMA_HeyServerHint:
	    attr.hey_server_hint = va_arg(args, const char *);
	    break;
	case FMA_StandardOut:
	    attr.out = va_arg(args, ostream *);
	    break;
	case FMA_StandardError:
	    attr.err = va_arg(args, ostream *);
	    break;
	default:
	    break;
	}
	tag = va_arg(args, int);
    }

    switch( op )
    {
    case FLO_QUERY:
	if( attr.hey_server_hint != NULL )
	{
	    if( strcmp(attr.hey_server_hint,
		       "IDL:BrokerPolicies/StrictPolicy:1.0") == 0 )
	    {
		retval = EXIT_SUCCESS;
	    }
	    else if( strcmp(attr.hey_server_hint,
			    "IDL:BrokerPolicies/WeightedPolicy:1.0") == 0 )
	    {
		retval = EXIT_SUCCESS;
	    }
	    else if( strcmp(attr.hey_server_hint,
			    "IDL:BrokerPolicies/PartitionPolicy:1.0") == 0 )
	    {
		retval = EXIT_SUCCESS;
	    }
	    else
	    {
	    }
	}
	else
	{
	    cout << "\tIDL:BrokerPolicies/StrictPolicy:1.0" << endl
		 << "\tIDL:BrokerPolicies/WeightedPolicy:1.0" << endl
		 << "\tIDL:BrokerPolicies/PartitionPolicy:1.0" << endl;
	}
	break;
    case FLO_HEY:
	if( attr.hey_server_hint != NULL )
	{
	    try
	    {
		if( strcmp(attr.hey_server_hint,
			   "IDL:BrokerPolicies/StrictPolicy:1.0") == 0 )
		{
		    retval = pfStrictPolicyHey(attr.orb, *attr.hey_parser);
		}
		else if( strcmp(attr.hey_server_hint,
				"IDL:BrokerPolicies/WeightedPolicy:1.0") == 0 )
		{
		    retval = pfWeightedPolicyHey(attr.orb, *attr.hey_parser);
		}
		else if( strcmp(attr.hey_server_hint,
				"IDL:BrokerPolicies/PartitionPolicy:1.0") == 0 )
		{
		    retval = pfPartitionPolicyHey(attr.orb, *attr.hey_parser);
		}
		else
		{
		}
	    }
	    catch(const HeyParserException &he)
	    {
		cerr << "Parse error: " << he << endl;
	    }
	    catch(const CORBA::SystemException &e)
	    {
		cerr << "Caught Exception: " << e << endl;
	    }
	    catch(...)
	    {
		cerr << "Caught an unhandled exception" << endl;
	    }
	}
	else
	{
	    retval = pfHey(attr.orb,
			   attr.poa,
			   *attr.hey_parser,
			   *attr.out,
			   *attr.err);
	}
	break;
    default:
	break;
    }
    
    return( retval );
}
